clear all;
close all;
clc

addpath('MY_LIB')

%%%%%% TV regularization %%%%%%

% scale of downsampling, integer number

rate = 2;
dr = rate;
dc = rate;
ds = rate;
d = dr*dc*ds;

%%
% generate 3D Shepp Logan phantom
 [gt,elli]=phantom3d(256);
 gt1 = gt./max(gt(:));

%% Guassian blurring kernel           
[mh,nh,th]=size(gt);
ml = mh/dr; nl = nh/dc; tl = th/ds;
m = ml*nl*tl;
blksz = [ml, nl, tl];
H = fspecial3('gaussian',9,3);

%% pad the blurring kernel

[m0,n0,t0]=size(H);
hpad=padarray(H,floor([mh-m0+1,nh-n0+1,th-t0+1]/2),'pre');
hpad=padarray(hpad,round([mh-m0-1,nh-n0-1,th-t0-1]/2),'post');

hp_c=fftshift(hpad);%centre the dc of the padded image

FB=fftn(hp_c);% Fourier transform of the padded blurring kernel 
FBC=conj(FB);% conjugate of the fourier transform of the padded blurring kernel
F2B=abs(FB).^2;

%% blur the image in the Fourier domain
Hx=real(ifftn(FB.*fftn(gt)));

%% add Gaussian noise
N= numel(gt);
BSNRdb = 30;
sigman = fro(Hx-mean(Hx(:)))/sqrt(N*10^(BSNRdb/10));
y1 = Hx + sigman*randn(size(gt));
FY = fftn(y1);

%% decimate the blurred/noisy image
dec=y1(1:dr:end,1:dc:end,1:ds:end);
dec1 =dec./max(dec(:));

%% upsample observed image by conjugate transpose decimation operator
decsize = size(dec);
decrup = decsize(1)*dr;
deccup = decsize(2)*dc;
decsup = decsize(3)*ds;
DTy=zeros(decrup,deccup,decsup);
DTy(1:dr:end,1:dc:end,1:ds:end)=dec;

%%  interpolation
tic;
[y,x,z]=...
    ndgrid(linspace(1, size(dec,1),decrup),...
           linspace(1, size(dec,2),deccup),...
           linspace(1, size(dec,3),decsup));...
           
decup_im = interp3(dec, x,y,z,'cubic');
decup_im1 = decup_im./max(decup_im(:));
toc;
psnr_in = calc_PSNR(gt,decup_im);
fprintf('interpo_psnr: PSNR = %g\n',psnr_in);


%% INITIALIZATION OF ADMM
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% define the difference operator kernel
dh = zeros(decrup,deccup,decsup);
dh(1,1) = 1;
dh(1,2) = -1;
% 
dv = zeros(decrup,deccup,decsup);
dv(1,1) = 1;
dv(2,1) = -1;
%

dss = zeros(decrup,deccup,decsup);
dss(1,1) = 1;
dss(2,1) = -1;
% 
% compute FFTs for filtering
FDH = fftn(dh);
FDHC = conj(FDH);
F2DH = abs(FDH).^2;
%
FDV = fftn(dv);
FDVC = conj(FDV);
F2DV = abs(FDV).^2;
%
FDS = fftn(dss);
FDSC = conj(FDS);
F2DS = abs(FDS).^2;
%
c = 1e-8;
F2D = F2DH + F2DV + F2DS +c;

STy = zeros(decrup,deccup,decsup);
STy(1:dr:end,1:dc:end,1:ds:end)=dec;

FBTSTy = FBC.*fftn(STy);

%% manually adjusted hyperparameters
stoppingrule = 1;
tolA = 1e-3;
sig2n = sigman^2;

%% manually adjusted hyperparameters
taup_list = 2.3333;
mu_list = 0.0042;

%%
maxiter = 200;
distance = zeros(3,maxiter);
distance_max = zeros(1,maxiter);
criterion = zeros(1,maxiter);
mses = zeros(1,maxiter);
objective = zeros(1,maxiter);

% findings
psnr_admm = zeros(1,maxiter);
isnr_admm = zeros(1,maxiter);
times = zeros(1,maxiter);


%% parameters
lam = taup_list*sig2n;
mu =  mu_list;
gam = lam./mu;
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
X1 = decup_im;
FX1 = fftn(X1);
BX1 = ifftn(FB.*fftn(X1));
resid1 =  dec - BX1(1:dr:end,1:dc:end,1:ds:end);
TVpenalty1 = sum(sum(sum(sqrt(abs(ifftn(FDH.*(FX1))).^2+abs(ifftn(FDV.*(FX1))).^2+abs(ifftn(FDS.*(FX1))).^2))));
prev_f = 0.5*(resid1(:)'*resid1(:)) + lam*TVpenalty1;
objective(:,1) = prev_f;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%
X = decup_im;
U1 = X;
U2 = X;
U3 = X;
    
D1 = 0*X;
D2 = D1;
D3 = D2;
    
t_start=  tic;
for j = 1:maxiter
   isnr_admm(1,j) = isnr_c(decup_im,gt,X); 
  
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %%%%%%%%%%%%%%%%%%%%%%%%%% update X %%%%%%%%%%%%%%%%%%%%%%%%%%%
   % argmin_x  .5*||y-SHx||^2 + 
   %           mu/2*||DhX - U1 + D1||^2 + 
   %           mu/2*||DvX - U2 + D2||^2 + 
   %           mu/2*||DsX - U3 + D3||^2
   
   V1 = U1-D1;
   V2 = U2-D2;
   V3 = U3-D3;
   
   FV1 = mu*FDHC.*fftn(V1);
   FV2 = mu*FDVC.*fftn(V2);
   FV3 = mu*FDSC.*fftn(V3);
   
   FR = FBTSTy + FV1 + FV2 + FV3;
   x11 = FB.*FR./F2D;
   fun = @(x) reshape(ufd_1(x,[1],[2,3]),m,1);
   FBR1 = blcthre3d(x11, blksz, fun);

   invW = blcthre3d(F2B./F2D, blksz, fun);
   invWBR = FBR1./(invW + (mu*d));
 
   fun2 = @(x) x.*invWBR;
   FCBinvWBR = blockproc3(FBC, blksz, fun2);
   FX = (FR-FCBinvWBR)./F2D/mu;
   Xest1 = real(ifftn(FX));
   Xest2 = Xest1./max(Xest1(:));
   
   %%%%%%%%%%%%%%%%%%%%%%%%%% STEP 1  update U %%%%%%%%%%%%%%%%%%%%%%%%%%%
   % argmin_u tau*||sqrt(U1^2 + U2^2 + U3^2)||_1 + 
   %                    mu/2*||DhX - U1 + D1||^2 +
   %                    mu/2*||DvX - U2 + D2||^2 +
   %                    mu/2*||DsX - U3 + D3||^2 
   DhX = ifftn(FX.*FDH);
   DvX = ifftn(FX.*FDV);
   DsX = ifftn(FX.*FDS);
   
   NU1 = DhX + D1;
   NU2 = DvX + D2;
   NU3 = DsX + D3;
   
   NU = sqrt(NU1.^2 + NU2.^2 + NU3.^2);
   A = max(0, NU-gam);
   %A = A./(A + gam(i));
   A = A./NU;% ideal from the solution
   U1 = A.*NU1;
   U2 = A.*NU2;
   U3 = A.*NU3;
   
    %%%%%%%%%%%%%%%%%%%%%%%%%% STEP 3 update D %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   D1 = D1 + (DhX-U1);
   D2 = D2 + (DvX-U2);
   D3 = D3 + (DsX-U3);
       
    %%
    BX = ifftn(FB.*FX);
    resid =  dec - BX(1:dr:end,1:dc:end,1:ds:end);
    TVpenalty = sum(sum(sum(sqrt(abs(ifftn(FDH.*(FX))).^2+abs(ifftn(FDV.*(FX))).^2+abs(ifftn(FDS.*(FX))).^2))));
    objective(1,j+1) = 0.5*(resid(:)'*resid(:)) + lam*TVpenalty;
    
    distance(1,j) = norm(DhX(:)-U1(:),2)^2;
    distance(2,j) = norm(DvX(:)-U2(:),2)^2;
    distance(3,j) = norm(DsX(:)-U3(:),2)^2;
    distance_max(1,j)=distance(1,j)+distance(2,j)+distance(3,j);
    err = Xest1-gt;
    mses(1,j) =  (err(:)'*err(:))/N;
     times(j) = +toc;
    
    switch stoppingrule
        case 1      
            criterion(1,j) = abs(objective(1,j+1)-objective(1,j))/objective(1,j);
        case 2
            criterion(1,j) = distance_max(1,j);
    end

     if ( criterion(1,j) < tolA )
         break
     end
    
     isnr_admm = isnr_c(decup_im,gt,Xest1);
        fprintf('FSR-ADMM_isnr: lam =  %g, mu = %g, Iter = %d, ISNR = %g\n',...
        lam, mu, j, isnr_admm);
    
     psnr_admm  = calc_PSNR(gt,Xest1);
        fprintf('FSR-ADMM_psnr: lam =  %g, mu = %g, Iter = %d, PSNR = %g\n',...
        lam, mu, j, psnr_admm);  
    
     times  = toc(t_start);
        fprintf('FSR-ADMM_time: lam =  %g, mu = %g, Iter = %d, Time = %g\n',...
        lam, mu, j, times);  
end

%%
figure;
imshow3D(gt1);
title('original image');
colormap gray; axis off

figure;
imshow3D(dec1);
title('decimated_blurred image');
colormap gray; axis off

figure;
imshow3D(decup_im1);
title('interpolated image');
colormap gray; axis off

figure;
imshow3D(Xest2);
title('FSR-ADMM');
colormap gray; axis off